# Copyright 2024 ByteDance and/or its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import dataclasses
import json
import logging
import os
import pickle
import random
import time
from datetime import datetime, timezone
from typing import Any, Dict, Mapping, Optional, Sequence, Tuple, Union

import numpy as np
from typing_extensions import Final, TypeAlias

from protenix.data.constants import (
    ATOM37_NUM,
    ATOM37_ORDER,
    PROTEIN_AATYPE_DENSE_ATOM_TO_ATOM37,
    PROTEIN_CHAIN,
    RESTYPE_PSEUDOBETA_INDEX,
    RESTYPE_RIGIDGROUP_DENSE_ATOM_IDX,
    STD_RESIDUES_WITH_GAP,
)
from protenix.data.template.template_parser import (
    AlignRatioError,
    CaDistanceError,
    DateError,
    DuplicateError,
    encode_template_restype,
    get_pdb_id_and_chain,
    LengthError,
    MmcifObject,
    MultipleChainsError,
    NoAtomDataInTemplateError,
    NoChainsError,
    PrefilterError,
    PrefilterResult,
    QueryToTemplateAlignError,
    SingleHitResult,
    TemplateAtomMaskAllZerosError,
    TemplateHit,
    TemplateParser,
    TemplateSearchResult,
)
from protenix.data.tools.kalign import Kalign
from protenix.utils.logger import get_logger

logger = get_logger(__name__)
logger.setLevel(logging.INFO)

# Global caches for release dates and obsolete PDBs to avoid redundant loading.
_RELEASE_DATES_CACHE: Dict[str, Mapping[str, datetime]] = {}
_OBSOLETE_PDBS_CACHE: Dict[str, Dict[str, str]] = {}

FeatureDict: TypeAlias = Mapping[str, np.ndarray]

DAYS_BEFORE_QUERY_DATE: Final[int] = 60

TEMPLATE_FEATURES: Final[Tuple[str, ...]] = (
    "template_aatype",
    "template_atom_positions",
    "template_atom_mask",
)

_POLYMER_FEATURES: Final[Mapping[str, Union[np.float64, np.int32, object]]] = {
    "template_aatype": np.int32,
    "template_all_atom_masks": np.float64,
    "template_all_atom_positions": np.float64,
    "template_domain_names": object,
    "template_release_date": object,
    "template_sequence": object,
}


@dataclasses.dataclass(frozen=True)
class DistogramFeaturesConfig:
    # The left edge of the first bin.
    min_bin: float = 3.25
    # The left edge of the final bin. The final bin catches everything larger than
    # `max_bin`.
    max_bin: float = 50.75
    # The number of bins in the distogram.
    num_bins: int = 39


class TemplateFeatures:
    """Utility functions for AlphaFold 3 template processing."""

    @staticmethod
    def get_timestamp(date_str: str) -> float:
        """Converts an ISO format date string to a UTC timestamp."""
        dt = datetime.fromisoformat(date_str)
        dt = dt.replace(tzinfo=timezone.utc)
        return dt.timestamp()

    @staticmethod
    def package_template_features(
        *, hit_features: Sequence[Mapping[str, Any]]
    ) -> Mapping[str, Any]:
        """Stacks polymer features, adds empty and keeps ligand features unstacked."""
        features_to_include = set(_POLYMER_FEATURES)
        features = {
            feat: [single_hit_features[feat] for single_hit_features in hit_features]
            for feat in features_to_include
        }

        stacked_features = {}
        for k, v in features.items():
            if k in _POLYMER_FEATURES:
                v = (
                    np.stack(v, axis=0)
                    if v
                    else np.array([], dtype=_POLYMER_FEATURES[k])
                )
            stacked_features[k] = v

        return stacked_features

    @staticmethod
    def fix_template_features(
        template_features: FeatureDict, num_res: int
    ) -> FeatureDict:
        """Convert template features to AlphaFold 3 format."""
        if not template_features["template_aatype"].shape[0]:
            template_features = TemplateFeatures.empty_template_features(num_res)
        else:
            template_release_timestamp = [
                TemplateFeatures.get_timestamp(x.decode("utf-8"))
                for x in template_features["template_release_date"]
            ]

            # Convert from atom37 to dense atom
            dense_atom_indices = np.take(
                PROTEIN_AATYPE_DENSE_ATOM_TO_ATOM37,
                template_features["template_aatype"],
                axis=0,
            )

            atom_mask = np.take_along_axis(
                template_features["template_all_atom_masks"], dense_atom_indices, axis=2
            )
            atom_positions = np.take_along_axis(
                template_features["template_all_atom_positions"],
                dense_atom_indices[..., None],
                axis=2,
            )
            atom_positions *= atom_mask[..., None]

            template_features = {
                "template_aatype": template_features["template_aatype"],
                "template_atom_mask": atom_mask.astype(np.int32),
                "template_atom_positions": atom_positions.astype(np.float32),
                "template_domain_names": np.array(
                    template_features["template_domain_names"], dtype=object
                ),
                "template_release_timestamp": np.array(
                    template_release_timestamp, dtype=np.float32
                ),
            }
        return template_features

    @staticmethod
    def empty_template_features(num_res: int, num_dense: int = 24) -> FeatureDict:
        """Creates fully masked out template features."""
        template_features = {
            "template_aatype": np.array(
                [STD_RESIDUES_WITH_GAP["-"]] * num_res, dtype=np.int32
            )[None, ...],
            "template_atom_mask": np.zeros((num_res, num_dense), dtype=np.int32)[
                None, ...
            ],
            "template_atom_positions": np.zeros(
                (num_res, num_dense, 3), dtype=np.float32
            )[None, ...],
            "template_domain_names": np.array([b""], dtype=object),
            "template_release_timestamp": np.array([0.0], dtype=np.float32),
        }
        return template_features

    @staticmethod
    def pseudo_beta_fn(
        aatype: np.ndarray,
        dense_atom_positions: np.ndarray,
        dense_atom_masks: np.ndarray,
        is_ligand: Optional[np.ndarray] = None,
    ) -> tuple[np.ndarray, np.ndarray]:
        """Create pseudo beta atom positions and optionally mask."""
        if is_ligand is None:
            is_ligand = np.zeros_like(aatype)

        pseudobeta_index_polymer = np.take(
            RESTYPE_PSEUDOBETA_INDEX, aatype, axis=0
        ).astype(np.int32)

        pseudobeta_index = np.where(
            is_ligand,
            np.zeros_like(pseudobeta_index_polymer),
            pseudobeta_index_polymer,
        )

        pseudo_beta = np.take_along_axis(
            dense_atom_positions, pseudobeta_index[..., None, None], axis=-2
        )
        pseudo_beta = np.squeeze(pseudo_beta, axis=-2)

        pseudo_beta_mask = np.take_along_axis(
            dense_atom_masks, pseudobeta_index[..., None], axis=-1
        ).astype(np.float32)
        pseudo_beta_mask = np.squeeze(pseudo_beta_mask, axis=-1)

        return pseudo_beta, pseudo_beta_mask

    @staticmethod
    def dgram_from_positions(
        positions: np.ndarray, config: DistogramFeaturesConfig
    ) -> np.ndarray:
        """Compute distogram from amino acid positions."""
        lower_breaks = np.linspace(config.min_bin, config.max_bin, config.num_bins)
        lower_breaks = np.square(lower_breaks)
        upper_breaks = np.concatenate(
            [lower_breaks[1:], np.array([1e8], dtype=np.float32)], axis=-1
        )
        dist2 = np.sum(
            np.square(
                np.expand_dims(positions, axis=-2) - np.expand_dims(positions, axis=-3)
            ),
            axis=-1,
            keepdims=True,
        )

        dgram = (dist2 > lower_breaks).astype(np.float32) * (
            dist2 < upper_breaks
        ).astype(np.float32)
        return dgram

    @staticmethod
    def compute_template_unit_vector(
        aatype: np.ndarray,
        atom_positions: np.ndarray,
        atom_mask: np.ndarray,
        epsilon: float = 1e-6,
    ) -> tuple[np.ndarray, np.ndarray]:
        """Simplified calculation of template unit vector."""
        # Get backbone indices (C, CA, N) for each residue from protein_data_processing
        # Group 0: [C, CA, N]
        backbone_indices = RESTYPE_RIGIDGROUP_DENSE_ATOM_IDX[aatype, 0]  # [num_res, 3]

        # Indices according to protein_data_processing.py: C is 0, CA is 1, N is 2
        c_idx = backbone_indices[:, 0]
        ca_idx = backbone_indices[:, 1]
        n_idx = backbone_indices[:, 2]

        num_res = aatype.shape[0]
        res_indices = np.arange(num_res)

        c_pos = atom_positions[res_indices, c_idx]
        ca_pos = atom_positions[res_indices, ca_idx]
        n_pos = atom_positions[res_indices, n_idx]

        c_mask = atom_mask[res_indices, c_idx]
        ca_mask = atom_mask[res_indices, ca_idx]
        n_mask = atom_mask[res_indices, n_idx]

        mask = (c_mask * ca_mask * n_mask).astype(np.float32)

        # Local frame: origin at CA
        # x-axis along C-CA (following original code convention)
        v1 = c_pos - ca_pos
        v2 = n_pos - ca_pos

        e1 = v1 / (np.linalg.norm(v1, axis=-1, keepdims=True) + epsilon)
        # Orthogonalize v2 against e1
        e2 = v2 - np.sum(v2 * e1, axis=-1, keepdims=True) * e1
        e2 = e2 / (np.linalg.norm(e2, axis=-1, keepdims=True) + epsilon)
        # e3 = e1 x e2
        e3 = np.cross(e1, e2)

        # Relative positions of all CA atoms to all local frames
        # diff[i, j] = CA[j] - CA[i]
        diff = ca_pos[None, :, :] - ca_pos[:, None, :]  # [num_res, num_res, 3]

        # Transform to local frame: P' = R^T @ diff
        # R = [e1 | e2 | e3] -> x' = e1 . diff, y' = e2 . diff, z' = e3 . diff
        ux = np.sum(e1[:, None, :] * diff, axis=-1)
        uy = np.sum(e2[:, None, :] * diff, axis=-1)
        uz = np.sum(e3[:, None, :] * diff, axis=-1)

        unit_vector = np.stack([ux, uy, uz], axis=-1)  # [num_res, num_res, 3]

        # Normalize to unit vector
        uv_norm = np.linalg.norm(unit_vector, axis=-1, keepdims=True)
        unit_vector = unit_vector / (uv_norm + epsilon)

        # 2D mask
        mask_2d = mask[:, None] * mask[None, :]

        return unit_vector, mask_2d


# --- Filtering and Processing ---


class TemplateHitFilter:
    """Filters template hits based on date, alignment ratio, and other criteria."""

    def __init__(
        self,
        release_dates: Mapping[str, datetime],
        obsolete_pdbs: Mapping[str, str],
        strict: bool = False,
    ):
        self.release_dates = release_dates
        self.obsolete_pdbs = obsolete_pdbs
        self.strict = strict

    def _is_after_cutoff(self, pdb_id: str, cutoff: Optional[datetime]) -> bool:
        """Checks if a PDB entry was released after the cutoff date."""
        if cutoff is None:
            return False
        date = self.release_dates.get(pdb_id)
        return date > cutoff if date else False

    def _assess_hit(
        self,
        hit: TemplateHit,
        pdb_code: str,
        query_seq: str,
        cutoff: datetime,
        max_subseq_ratio: float = 0.95,
        min_align_ratio: float = 0.1,
    ) -> bool:
        """Performs quick pre-filtering on a hit."""
        align_ratio = hit.aligned_cols / len(query_seq)
        t_seq = hit.hit_sequence.replace("-", "")
        len_ratio = len(t_seq) / len(query_seq)

        if self._is_after_cutoff(pdb_code, cutoff):
            raise DateError(f"Release date for {pdb_code} is after cutoff.")
        if align_ratio <= min_align_ratio:
            raise AlignRatioError(f"Align ratio {align_ratio:.2f} <= {min_align_ratio}")
        if t_seq in query_seq and len_ratio > max_subseq_ratio:
            raise DuplicateError("Hit is a large duplicate of the query.")
        if len(t_seq) < 10:
            raise LengthError("Template sequence too short.")
        return True

    def prefilter(
        self,
        query_seq: str,
        hit: TemplateHit,
        max_date: datetime,
    ) -> PrefilterResult:
        """Prefilters a hit and handles obsolete PDBs."""
        try:
            pdb_id, chain_id = get_pdb_id_and_chain(hit)
        except ValueError as e:
            return PrefilterResult(False, error=str(e))

        if pdb_id not in self.release_dates:
            pdb_id = self.obsolete_pdbs.get(pdb_id, pdb_id)
            if pdb_id not in self.release_dates:
                return PrefilterResult(False, error=f"{pdb_id} date unknown.")

        try:
            self._assess_hit(hit, pdb_id, query_seq, max_date)
        except PrefilterError as e:
            msg = f"Hit {pdb_id}_{chain_id} failed prefilter: {e}"
            if self.strict and isinstance(e, (DateError, DuplicateError)):
                return PrefilterResult(False, error=msg)
            return PrefilterResult(False, warning=msg)
        return PrefilterResult(True)


class TemplateHitProcessor:
    """Processes a template hit to generate features."""

    def __init__(
        self,
        mmcif_dir: str,
        template_cache_dir: Optional[str] = None,
        kalign_binary_path: Optional[str] = None,
        _zero_center_positions: bool = True,
    ):
        self._mmcif_dir = mmcif_dir
        self._template_cache_dir = template_cache_dir
        self._kalign_binary_path = kalign_binary_path
        self._zero_center_positions = _zero_center_positions

    def _read_file(self, path: str) -> str:
        """Reads file content."""
        with open(path, "r") as f:
            return f.read()

    def _get_atom_coords(
        self,
        mmcif_object: MmcifObject,
        chain_id: str,
        _zero_center_positions: bool = False,
    ) -> Tuple[np.ndarray, np.ndarray]:
        """Extracts all-atom coordinates and mask for a specific chain."""
        chains = list(mmcif_object.structure.get_chains())
        relevant_chains = [c for c in chains if c.id == chain_id]
        if len(relevant_chains) != 1:
            raise MultipleChainsError(
                f"Expected 1 chain with id {chain_id}, found {len(relevant_chains)}"
            )
        chain = relevant_chains[0]

        num_res = len(mmcif_object.chain_to_seqres[chain_id])
        all_pos = np.zeros((num_res, ATOM37_NUM, 3), dtype=np.float32)
        all_mask = np.zeros((num_res, ATOM37_NUM), dtype=np.float32)

        for i in range(num_res):
            res_info = mmcif_object.seqres_to_structure[chain_id][i]
            if not res_info.is_missing:
                try:
                    res = chain[
                        (
                            res_info.hetflag,
                            res_info.position.residue_number,
                            res_info.position.insertion_code,
                        )
                    ]
                    for atom in res.get_atoms():
                        name = atom.get_name()
                        coord = atom.get_coord()
                        if name in ATOM37_ORDER:
                            idx = ATOM37_ORDER[name]
                            all_pos[i, idx] = coord
                            all_mask[i, idx] = 1.0
                        elif name.upper() == "SE" and res.get_resname() == "MSE":
                            idx = ATOM37_ORDER["SD"]
                            all_pos[i, idx] = coord
                            all_mask[i, idx] = 1.0

                    # Correct Arginine NH1/NH2 if swapped based on distance to CD.
                    cd, nh1, nh2 = (
                        ATOM37_ORDER["CD"],
                        ATOM37_ORDER["NH1"],
                        ATOM37_ORDER["NH2"],
                    )
                    if res.get_resname() == "ARG" and all(all_mask[i, [cd, nh1, nh2]]):
                        if np.linalg.norm(
                            all_pos[i, nh1] - all_pos[i, cd]
                        ) > np.linalg.norm(all_pos[i, nh2] - all_pos[i, cd]):
                            all_pos[i, [nh1, nh2]] = all_pos[i, [nh2, nh1]]
                            all_mask[i, [nh1, nh2]] = all_mask[i, [nh2, nh1]]
                except KeyError:
                    continue

        if _zero_center_positions:
            mask_bool = all_mask.astype(bool)
            if np.any(mask_bool):
                center = all_pos[mask_bool].mean(axis=0)
                all_pos[mask_bool] -= center

        return all_pos, all_mask

    def _check_residue_distances(
        self, pos: np.ndarray, mask: np.ndarray, max_dist: float
    ):
        """Verifies that distance between consecutive CA atoms is within limits."""
        ca_idx = ATOM37_ORDER["CA"]
        prev_pos = None
        for i, (p, m) in enumerate(zip(pos, mask)):
            if m[ca_idx]:
                curr_pos = p[ca_idx]
                if prev_pos is not None:
                    dist = np.linalg.norm(curr_pos - prev_pos)
                    if dist > max_dist:
                        raise CaDistanceError(
                            f"Distance between residues {i} and previous is {dist:.2f} > {max_dist}"
                        )
                prev_pos = curr_pos

    def _get_atom_positions(
        self,
        mmcif_obj: MmcifObject,
        auth_chain_id: str,
        max_ca_dist: float,
        _zero_center: bool = False,
    ) -> Tuple[np.ndarray, np.ndarray]:
        pos, mask = self._get_atom_coords(mmcif_obj, auth_chain_id, _zero_center)
        self._check_residue_distances(pos, mask, max_ca_dist)
        return pos, mask

    def _align_query_to_hit_index_mapping(
        self,
        query_seq: str,
        chain_id: str,
        mmcif_obj: MmcifObject,
        min_ratio: float,
    ) -> Tuple[str, Dict[int, int], str]:
        """
        Aligns the query sequence directly to the chain sequence in mmCIF and establishes index mapping.

        This method ensures that each residue in the query sequence finds its corresponding
        position in the template structure sequence via alignment.

        Args:
            query_seq: The query sequence string.
            chain_id: The ID of the target chain.
            mmcif_obj: The parsed mmCIF object.
            min_ratio: Minimum allowed sequence identity ratio.

        Returns:
            target_seq: The original chain sequence from mmCIF.
            mapping: A dictionary establishing the mapping from query sequence index to
                    template sequence index (Query Index -> Template Index).
                    key: Residue index in the query sequence (0-based).
                    value: Residue index in the template sequence (0-based).
            actual_chain_id: The actual chain ID used from the mmCIF.
        """
        aligner = Kalign(binary_path=self._kalign_binary_path)
        target_seq = mmcif_obj.chain_to_seqres.get(chain_id, "")
        actual_chain_id = chain_id
        if not target_seq:
            if len(mmcif_obj.chain_to_seqres) == 1:
                actual_chain_id = list(mmcif_obj.chain_to_seqres.keys())[0]
                target_seq = mmcif_obj.chain_to_seqres[actual_chain_id]
            else:
                raise QueryToTemplateAlignError(f"Chain {chain_id} not found.")

        try:
            q_aln, t_aln = aligner.align([query_seq, target_seq])
        except Exception as e:
            raise QueryToTemplateAlignError(f"Query alignment failed: {e}")

        mapping = {}
        q_idx, t_idx, same, count = -1, -1, 0, 0
        for qa, ta in zip(q_aln, t_aln):
            if qa != "-":
                q_idx += 1
            if ta != "-":
                t_idx += 1
            if qa != "-" and ta != "-":
                # Establish mapping from Query index to Template (mmCIF) index
                mapping[q_idx] = t_idx
                count += 1
                if qa == ta:
                    same += 1

        if count > 0 and (same / count < min_ratio):
            raise QueryToTemplateAlignError("Insufficient similarity to query.")

        return target_seq, mapping, actual_chain_id

    def _extract_template_features(
        self,
        mmcif_obj: MmcifObject,
        pdb_id: str,
        mapping: Mapping[int, int],
        template_seq: str,
        query_seq: str,
        chain_id: str,
        _zero_center: bool = True,
    ) -> Tuple[Dict[str, Any], Optional[str]]:
        """Generates features for a template hit."""
        if not mmcif_obj or not mmcif_obj.chain_to_seqres:
            raise NoChainsError(f"No chains in {pdb_id}")

        warning = None
        try:
            all_pos, all_mask = self._get_atom_positions(
                mmcif_obj, chain_id, 150.0, _zero_center
            )
        except (CaDistanceError, KeyError) as e:
            raise NoAtomDataInTemplateError(f"Failed to get atom data: {e}")

        num_query = len(query_seq)
        out_pos = np.zeros((num_query, ATOM37_NUM, 3), dtype=np.float32)
        out_mask = np.zeros((num_query, ATOM37_NUM), dtype=np.float32)
        out_seq = ["-"] * num_query

        for q_idx, t_idx in mapping.items():
            if t_idx != -1:
                out_pos[q_idx] = all_pos[t_idx]
                out_mask[q_idx] = all_mask[t_idx]
                out_seq[q_idx] = template_seq[t_idx]

        if np.sum(out_mask) < 5:
            raise TemplateAtomMaskAllZerosError(
                f"Empty atom mask for {pdb_id}_{chain_id}"
            )

        out_seq_str = "".join(out_seq)
        aatype = encode_template_restype(PROTEIN_CHAIN, out_seq_str)

        features = {
            "template_all_atom_positions": out_pos,
            "template_all_atom_masks": out_mask,
            "template_sequence": out_seq_str.encode(),
            "template_aatype": np.array(aatype, dtype=np.int32),
            "template_domain_names": np.array(
                f"{pdb_id.lower()}_{chain_id}".encode(), dtype=object
            ),
        }
        return features, warning

    def _update_realigned_hit(
        self, hit: TemplateHit, seq: str, mapping: Mapping[int, int]
    ) -> TemplateHit:
        """
        Updates the TemplateHit object with new alignment information.

        This method synchronizes the hit information (alignment columns, hit sequence,
        and index mappings) after a re-alignment between the query and the template structure.

        Args:
            hit: The original TemplateHit object.
            seq: The new hit sequence. Note: This sequence MUST NOT contain GAPs ('-').
            mapping: A dictionary mapping Query Index -> Template Index.

        Returns:
            A new TemplateHit object with updated alignment details.
        """
        q_indices = list(range(len(hit.query)))
        h_indices = [-1] * len(hit.query)
        for i, j in mapping.items():
            h_indices[i] = j
        return TemplateHit(
            index=hit.index,
            name=hit.name,
            aligned_cols=len(mapping),
            sum_probs=hit.sum_probs,
            query=hit.query,
            hit_sequence=seq,  # GAP-free sequence
            indices_query=q_indices,
            indices_hit=h_indices,
        )

    def process(
        self,
        query_seq: str,
        hit: TemplateHit,
        max_date: datetime,
        release_dates: Mapping[str, datetime],
        obsolete_pdbs: Mapping[str, str],
        strict: bool = False,
    ) -> Tuple[SingleHitResult, Dict[str, float]]:
        """Processes a single hit to generate features."""
        track = {}
        t_start = time.time()
        pdb_id, chain_id = get_pdb_id_and_chain(hit)
        pdb_id = obsolete_pdbs.get(pdb_id, pdb_id)

        # Load mmCIF parsing result.
        if self._template_cache_dir:
            cache_path = os.path.join(
                self._template_cache_dir, f"{pdb_id}_{chain_id}.pkl"
            )
            if os.path.exists(cache_path):
                with open(cache_path, "rb") as f:
                    res = pickle.load(f)
                track["load"] = time.time() - t_start
            else:
                return SingleHitResult(None, None, None, None), track
        else:
            # The raw mmCIF file can be split into per-chain files named
            # {pdb_id}_{chain_id}.cif to accelerate the loading process.
            cif_path = os.path.join(self._mmcif_dir, f"{pdb_id}.cif")
            try:
                cif_str = self._read_file(cif_path)
                res = TemplateParser.parse(
                    file_id=pdb_id, mmcif_string=cif_str, auth_chain_id=chain_id
                )
                track["load"] = time.time() - t_start
            except FileNotFoundError:
                return (
                    SingleHitResult(None, None, f"CIF not found: {cif_path}", None),
                    track,
                )

        if not res.mmcif_object:
            return (
                SingleHitResult(None, None, f"Parsing failed for {pdb_id}", None),
                track,
            )

        date_str = res.mmcif_object.header.get("release_date", "9999-12-31")
        hit_date = datetime.strptime(date_str, "%Y-%m-%d")
        if hit_date > max_date:
            err = f"Hit {pdb_id} date {hit_date} > {max_date}"
            return (
                SingleHitResult(None, None, err if strict else None, None),
                track,
            )

        # Re-align and extract features.
        t_aln = time.time()
        try:
            (
                target_seq,
                mapping,
                actual_chain_id,
            ) = self._align_query_to_hit_index_mapping(
                query_seq, chain_id, res.mmcif_object, 0.0
            )
            hit = self._update_realigned_hit(hit, target_seq, mapping)
            track["align"] = time.time() - t_aln

            t_feat = time.time()
            feats, warn = self._extract_template_features(
                res.mmcif_object,
                pdb_id,
                mapping,
                target_seq,
                query_seq,
                actual_chain_id,
                self._zero_center_positions,
            )
            feats["template_sum_probs"] = [
                hit.sum_probs if hit.sum_probs is not None else 0.0
            ]
            feats["template_release_date"] = np.array(date_str.encode(), dtype=object)
            track["feature"] = time.time() - t_feat
            return SingleHitResult(hit, feats, None, warn), track
        except Exception as e:
            return (
                SingleHitResult(None, None, f"Error processing hit: {e}", None),
                track,
            )


class TemplateHitFeaturizer:
    """
    Featurizes template hits (e.g., from hmmsearch or HHSearch) into features.

    Args:
        mmcif_dir: Directory containing mmCIF files.
        template_cache_dir: Optional directory to cache parsed mmCIF objects.
        max_hits: Maximum number of hits to process.
        kalign_binary_path: Path to the kalign binary.
        max_template_date: Optional cutoff date for templates.
        release_dates_path: Path to the JSON file containing PDB release dates.
        obsolete_pdbs_path: Path to the JSON file containing obsolete PDB mappings.
        strict_error_check: If True, raise errors for filtering failures.
        _shuffle_top_k_prefiltered: If set, shuffle the top K hits after pre-filtering.
        _zero_center_positions: If True, center atom positions at zero.
        _max_template_candidates_num: Maximum number of candidate hits to consider.
    """

    def __init__(
        self,
        mmcif_dir: str,
        template_cache_dir: Optional[str] = None,
        max_hits: Optional[int] = None,
        kalign_binary_path: Optional[str] = None,
        max_template_date: Optional[Union[str, datetime]] = None,
        release_dates_path: Optional[str] = None,
        obsolete_pdbs_path: Optional[str] = None,
        strict_error_check: bool = False,
        _shuffle_top_k_prefiltered: Optional[int] = None,
        _zero_center_positions: bool = True,
        _max_template_candidates_num: Optional[int] = None,
    ):
        self._mmcif_dir = mmcif_dir
        self._template_cache_dir = template_cache_dir
        self._max_hits = max_hits
        self._kalign_binary_path = kalign_binary_path
        self._strict_error_check = strict_error_check
        self._shuffle_top_k_prefiltered = _shuffle_top_k_prefiltered
        self._zero_center_positions = _zero_center_positions
        self._max_template_candidates_num = _max_template_candidates_num

        if max_template_date:
            if isinstance(max_template_date, str):
                self._max_template_date = datetime.strptime(
                    max_template_date, "%Y-%m-%d"
                )
            else:
                self._max_template_date = max_template_date
        else:
            self._max_template_date = None

        # Load or reuse release dates.
        if release_dates_path:
            if release_dates_path not in _RELEASE_DATES_CACHE:
                _RELEASE_DATES_CACHE[release_dates_path] = self._parse_release_dates(
                    release_dates_path
                )
            self._release_dates = _RELEASE_DATES_CACHE[release_dates_path]
        else:
            self._release_dates = {}

        # Load or reuse obsolete PDB mapping.
        if obsolete_pdbs_path and os.path.exists(obsolete_pdbs_path):
            if obsolete_pdbs_path not in _OBSOLETE_PDBS_CACHE:
                with open(obsolete_pdbs_path, "r") as f:
                    _OBSOLETE_PDBS_CACHE[obsolete_pdbs_path] = json.load(f)
            self._obsolete_pdbs = _OBSOLETE_PDBS_CACHE[obsolete_pdbs_path]
        else:
            self._obsolete_pdbs = {}

        self._hit_filter = TemplateHitFilter(
            release_dates=self._release_dates,
            obsolete_pdbs=self._obsolete_pdbs,
            strict=self._strict_error_check,
        )

        self._hit_processor = TemplateHitProcessor(
            mmcif_dir=self._mmcif_dir,
            template_cache_dir=self._template_cache_dir,
            kalign_binary_path=self._kalign_binary_path,
            _zero_center_positions=self._zero_center_positions,
        )

    def _parse_release_dates(self, path: str) -> Dict[str, datetime]:
        if not os.path.exists(path):
            return {}
        with open(path, "r") as f:
            data = json.load(f)
        return {
            pdb.lower(): datetime.strptime(v["release_date"], "%Y-%m-%d")
            for pdb, v in data.items()
            if "release_date" in v
        }

    def get_templates(
        self,
        sequence_uid: str,
        query_sequence: str,
        hits: Sequence[TemplateHit],
        max_template_date: Optional[Union[str, datetime]] = None,
    ) -> Tuple[TemplateSearchResult, Dict[str, float]]:
        """
        Processes hits to generate template features.

        Args:
            sequence_uid: Unique identifier for the sequence.
            query_sequence: The query sequence.
            hits: A sequence of TemplateHit objects.
            max_template_date: Optional cutoff date to override the default.

        Returns:
            A tuple of (TemplateSearchResult, timing_dict).
        """
        cutoff = self._max_template_date
        if max_template_date:
            if isinstance(max_template_date, str):
                cutoff = datetime.strptime(max_template_date, "%Y-%m-%d")
            else:
                cutoff = max_template_date

        # Prefilter hits.
        valid_hits = []
        errors, warnings = [], []
        for hit in hits:
            res = self._hit_filter.prefilter(
                query_sequence,
                hit,
                cutoff,
            )
            if res.valid:
                valid_hits.append(hit)
            if res.error:
                errors.append(res.error)
            if res.warning:
                warnings.append(res.warning)

        # Sort hits by sum_probs.
        valid_hits.sort(
            key=lambda x: x.sum_probs if x.sum_probs is not None else 0.0, reverse=True
        )

        # De-duplicate by hit sequence.
        deduped, seen_seq = [], set()
        for hit in valid_hits:
            seq = hit.hit_sequence.replace("-", "")
            if seq not in seen_seq:
                deduped.append(hit)
                seen_seq.add(seq)

        indices = list(range(len(deduped)))
        if self._shuffle_top_k_prefiltered:
            k = self._shuffle_top_k_prefiltered
            indices[:k] = np.random.permutation(indices[:k])

        if self._max_template_candidates_num:
            indices = indices[: self._max_template_candidates_num]

        # Determine number of templates to collect.
        max_to_collect = (
            self._max_hits
            if self._max_hits is not None
            else min(4, random.randint(0, min(len(indices), 20)))
        )

        features, final_hits = [], []
        already_seen_seqs = set()
        last_track = {}

        for i in indices:
            if len(features) >= max_to_collect:
                break
            hit = deduped[i]
            res, track = self._hit_processor.process(
                query_sequence,
                hit,
                cutoff,
                self._release_dates,
                self._obsolete_pdbs,
                self._strict_error_check,
            )
            last_track = track
            if res.error:
                errors.append(res.error)
            if res.warning:
                warnings.append(res.warning)
            if res.features:
                seq_key = res.features["template_sequence"]
                if seq_key not in already_seen_seqs:
                    already_seen_seqs.add(seq_key)
                    features.append(res.features)
                    final_hits.append(res.hit)

        return TemplateSearchResult(features, final_hits, errors, warnings), last_track



# Copyright 2024 ByteDance and/or its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
import tarfile
import time
from typing import Dict, List, Tuple

import requests
from requests.auth import HTTPBasicAuth
from tqdm import tqdm

TQDM_BAR_FORMAT = "{l_bar}{bar}| {n_fmt}/{total_fmt} [elapsed: {elapsed} estimate remaining: {remaining}]"
logger = logging.getLogger(__name__)

username = "example_user"
password = "example_password"


def parse_fasta_string(fasta_string: str) -> Dict:
    fasta_dict = {}
    lines = fasta_string.strip().split("\n")
    for line in lines:
        if line.startswith(">"):
            header = line[1:].strip()
            fasta_dict[header] = ""
        else:
            fasta_dict[header] += line.strip()
    return fasta_dict


def run_mmseqs2_service(
    x,
    prefix,
    use_env=True,
    use_filter=True,
    use_templates=False,
    filter=None,
    use_pairing=False,
    pairing_strategy="complete",
    host_url="https://api.colabfold.com",
    user_agent: str = "",
    email: str = "",
    server_mode: str = "protenix",
) -> Tuple[List[str], List[str]]:
    if server_mode == "protenix":
        assert host_url == "https://protenix-server.com/api/msa"
    submission_endpoint = "ticket/pair" if use_pairing else "ticket/msa"
    headers = {}
    if user_agent != "":
        headers["User-Agent"] = user_agent
    else:
        logger.warning(
            "No user agent specified. Please set a user agent (e.g., 'toolname/version contact@email') to help us debug in case of problems. This warning will become an error in the future."
        )

    def submit(seqs, mode, N=101):
        n, query = N, ""
        for seq in seqs:
            query += f"{seq}\n"
            n += 1

        while True:
            error_count = 0
            try:
                # https://requests.readthedocs.io/en/latest/user/advanced/#advanced
                # "good practice to set connect timeouts to slightly larger than a multiple of 3"
                res = requests.post(
                    f"{host_url}/{submission_endpoint}",
                    data={"q": query, "mode": mode, "email": email},
                    timeout=6.02,
                    headers=headers,
                    auth=HTTPBasicAuth(username, password),
                )
            except requests.exceptions.Timeout:
                logger.warning("Timeout while submitting to MSA server. Retrying...")
                continue
            except Exception as e:
                error_count += 1
                logger.warning(
                    f"Error while fetching result from MSA server. Retrying... ({error_count}/5)"
                )
                logger.warning(f"Error: {e}")
                time.sleep(5)
                if error_count > 5:
                    raise
                continue
            break

        try:
            out = res.json()
        except ValueError:
            logger.error(f"Server didn't reply with json: {res.text}")
            out = {"status": "ERROR"}
        return out

    def status(ID):
        while True:
            error_count = 0
            try:
                res = requests.get(
                    f"{host_url}/ticket/{ID}",
                    timeout=6.02,
                    headers=headers,
                    auth=HTTPBasicAuth(username, password),
                )
            except requests.exceptions.Timeout:
                logger.warning(
                    "Timeout while fetching status from MSA server. Retrying..."
                )
                continue
            except Exception as e:
                error_count += 1
                logger.warning(
                    f"Error while fetching result from MSA server. Retrying... ({error_count}/5)"
                )
                logger.warning(f"Error: {e}")
                time.sleep(5)
                if error_count > 5:
                    raise
                continue
            break
        try:
            out = res.json()
        except ValueError:
            logger.error(f"Server didn't reply with json: {res.text}")
            out = {"status": "ERROR"}
        return out

    def download(ID, path):
        error_count = 0
        while True:
            try:
                res = requests.get(
                    f"{host_url}/result/download/{ID}",
                    timeout=6.02,
                    headers=headers,
                    auth=HTTPBasicAuth(username, password),
                )
            except requests.exceptions.Timeout:
                logger.warning(
                    "Timeout while fetching result from MSA server. Retrying..."
                )
                continue
            except Exception as e:
                error_count += 1
                logger.warning(
                    f"Error while fetching result from MSA server. Retrying... ({error_count}/5)"
                )
                logger.warning(f"Error: {e}")
                time.sleep(5)
                if error_count > 5:
                    raise
                continue
            break
        with open(path, "wb") as out:
            out.write(res.content)

    # process input x
    seqs = [x] if isinstance(x, str) else x

    # compatibility to old option
    if filter is not None:
        use_filter = filter

    # setup mode
    if use_filter:
        mode = "env" if use_env else "all"
    else:
        mode = "env-nofilter" if use_env else "nofilter"

    if use_pairing:
        use_templates = False
        use_env = False
        mode = ""
        # greedy is default, complete was the previous behavior
        if pairing_strategy == "greedy":
            mode = "pairgreedy"
        elif pairing_strategy == "complete":
            mode = "paircomplete"

    # define path
    path = prefix
    if not os.path.isdir(path):
        os.mkdir(path)

    # call mmseqs2 api
    tar_gz_file = f"{path}/out.tar.gz"
    N, REDO = 101, True

    # deduplicate and keep track of order
    seqs_unique = []
    # TODO this might be slow for large sets
    [seqs_unique.append(x) for x in seqs if x not in seqs_unique]
    Ms = [N + seqs_unique.index(seq) for seq in seqs]
    # lets do it!
    logger.info("Msa server is running.")
    if not os.path.isfile(tar_gz_file):
        TIME_ESTIMATE = 100
        with tqdm(total=TIME_ESTIMATE, bar_format=TQDM_BAR_FORMAT) as pbar:
            while REDO:
                pbar.set_description("SUBMIT")

                # Resubmit job until it goes through
                out = submit(seqs_unique, mode, N)
                while out["status"] in ["UNKNOWN", "RATELIMIT"]:
                    sleep_time = 60
                    logger.error(f"Sleeping for {sleep_time}s. Reason: {out['status']}")
                    # resubmit
                    time.sleep(sleep_time)
                    out = submit(seqs_unique, mode, N)

                if out["status"] == "ERROR":
                    raise Exception(
                        "MMseqs2 API is giving errors. Please confirm your input is a valid protein sequence. If error persists, please try again an hour later."
                    )

                if out["status"] == "MAINTENANCE":
                    raise Exception(
                        "MMseqs2 API is undergoing maintenance. Please try again in a few minutes."
                    )

                # wait for job to finish
                ID, TIME = out["id"], 0
                pbar.set_description(out["status"])
                while out["status"] in ["UNKNOWN", "RUNNING", "PENDING"]:
                    t = 60
                    logger.error(f"Sleeping for {t}s. Reason: {out['status']}")
                    time.sleep(t)
                    out = status(ID)
                    pbar.set_description(out["status"])
                    if out["status"] == "RUNNING":
                        TIME += t
                    pbar.n = min(99, int(100 * TIME / (30.0 * 60)))
                    pbar.refresh()
                if out["status"] == "COMPLETE":
                    pbar.n = 100
                    pbar.refresh()
                    REDO = False

                if out["status"] == "ERROR":
                    REDO = False
                    raise Exception(
                        "MMseqs2 API is giving errors. Please confirm your input is a valid protein sequence. If error persists, please try again an hour later."
                    )

            # Download results
            download(ID, tar_gz_file)
            with tarfile.open(tar_gz_file) as tar_gz:
                tar_gz.extractall(os.path.dirname(tar_gz_file))
            files = os.listdir(os.path.dirname(tar_gz_file))

            if server_mode == "protenix":
                if (
                    "0.a3m" not in files
                    or "pdb70_220313_db.m8" not in files
                    or "uniref_tax.m8" not in files
                ):
                    raise FileNotFoundError(
                        "Files 0.a3m, pdb70_220313_db.m8, and uniref_tax.m8 not found in the directory."
                    )
                else:
                    print("Files downloaded and extracted successfully.")
            elif server_mode == "colabfold":
                if not use_pairing:
                    env_a3m_fpath = os.path.join(
                        prefix, "bfd.mgnify30.metaeuk30.smag30.a3m"
                    )
                    env_a3m_dict = parse_fasta_string(
                        open(env_a3m_fpath, "r").read().replace("\x00", "")
                    )
                    uniref_a3m_fpath = os.path.join(prefix, "uniref.a3m")
                    uniref_a3m_dict = parse_fasta_string(
                        open(uniref_a3m_fpath, "r").read().replace("\x00", "")
                    )
                    query_id = str(int(x.split("\n")[0].split("_")[-1]))
                    query_seq = x.split("\n")[1]
                    real_non_pairing_fpath = os.path.join(
                        prefix,
                        query_id,
                        "non_pairing.a3m",
                    )
                    output_dir = os.path.dirname(real_non_pairing_fpath)
                    if not os.path.exists(output_dir):
                        os.makedirs(output_dir)
                    with open(real_non_pairing_fpath, "w") as f:
                        f.write(f">query\n{query_seq}\n")
                        for k, v in env_a3m_dict.items():
                            if k.startswith("query_"):
                                continue
                            else:
                                f.write(f">{k}\n{v}\n")
                        for k, v in uniref_a3m_dict.items():
                            if k.startswith("query_"):
                                continue
                            else:
                                f.write(f">{k}\n{v}\n")
                    return os.path.abspath(os.path.dirname(real_non_pairing_fpath))
                else:
                    # pairing mode
                    pair_a3m = os.path.join(prefix, "pair.a3m")
                    pair_a3m_chunks = open(pair_a3m, "r").read().split("\x00")
                    for chunk in pair_a3m_chunks[:-1]:
                        real_pairing_fpath = os.path.join(
                            prefix,
                            str(int(chunk.split("\n")[0].split("_")[-1])),
                            "pairing.a3m",
                        )
                        output_dir = os.path.dirname(real_pairing_fpath)
                        if not os.path.exists(output_dir):
                            os.makedirs(output_dir)
                        chunk_fasta = parse_fasta_string(chunk)
                        with open(real_pairing_fpath, "w") as f:
                            for i, (k, v) in enumerate(chunk_fasta.items()):
                                if k.startswith("query_"):
                                    f.write(f">query\n{v}\n")
                                else:
                                    ks = k.split("\t")
                                    ks[0] = f"{ks[0]}_{i}/"
                                    k = "\t".join(ks)
                                    f.write(f">{k}_{i}\n{v}\n")



# Copyright 2024 ByteDance and/or its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import logging
import os
import time
import traceback
import warnings
from typing import Any, Mapping

import torch
from biotite.structure import AtomArray
from torch.utils.data import DataLoader, Dataset, DistributedSampler

from protenix.data.esm.esm_featurizer import ESMFeaturizer
from protenix.data.inference.json_to_feature import SampleDictToFeatures
from protenix.data.msa.msa_featurizer import InferenceMSAFeaturizer
from protenix.data.template.template_featurizer import InferenceTemplateFeaturizer
from protenix.data.template.template_utils import TemplateHitFeaturizer
from protenix.data.utils import data_type_transform, make_dummy_feature
from protenix.utils.distributed import DIST_WRAPPER
from protenix.utils.torch_utils import collate_fn_identity, dict_to_tensor

logger = logging.getLogger(__name__)

warnings.filterwarnings("ignore", module="biotite")


def get_inference_dataloader(configs: Any) -> DataLoader:
    """
    Creates and returns a DataLoader for inference using the InferenceDataset.

    Args:
        configs: A configuration object containing the necessary parameters for the DataLoader.

    Returns:
        A DataLoader object configured for inference.
    """
    inference_dataset = InferenceDataset(
        configs=configs,
    )
    sampler = DistributedSampler(
        dataset=inference_dataset,
        num_replicas=DIST_WRAPPER.world_size,
        rank=DIST_WRAPPER.rank,
        shuffle=False,
    )
    dataloader = DataLoader(
        dataset=inference_dataset,
        batch_size=1,
        sampler=sampler,
        collate_fn=collate_fn_identity,
        num_workers=configs.num_workers,
    )
    return dataloader


class InferenceDataset(Dataset):
    def __init__(
        self,
        configs,
    ) -> None:
        self.configs = configs

        self.input_json_path = configs.input_json_path
        self.dump_dir = configs.dump_dir
        self.use_msa = configs.use_msa
        self.msa_pair_as_unpair = configs.get("msa_pair_as_unpair", True)
        self.use_rna_msa = configs.get("use_rna_msa", True)
        self.use_template = configs.get("use_template", True)
        with open(self.input_json_path, "r") as f:
            self.inputs = json.load(f)
        json_task_name = os.path.basename(self.input_json_path).split(".")[0]
        if self.use_template:
            template_mmcif_dir = configs.data.template.prot_template_mmcif_dir
            assert template_mmcif_dir is not None and os.path.exists(
                template_mmcif_dir
            ), (
                "Inference with template depends on the mmcif directory.\n"
                "The mmcif directory containing cif files should be placed under $PROTENIX_ROOT_DIR/mmcif.\n"
                "You can download it from PDB https://www.wwpdb.org/ftp/pdb-ftp-sites or\n"
                "refer to scripts/database/download_protenix_data.sh to download inference dependency files, "
                "or set use_template=false for inference."
            )
            self.online_template_featurizer = TemplateHitFeaturizer(
                mmcif_dir=configs.data.template.prot_template_mmcif_dir,
                template_cache_dir=configs.data.template.prot_template_cache_dir,
                max_hits=4,
                kalign_binary_path=configs.data.template.kalign_binary_path,
                max_template_date="2021-09-30",
                release_dates_path=configs.data.template.release_dates_path,
                obsolete_pdbs_path=configs.data.template.obsolete_pdbs_path,
                _shuffle_top_k_prefiltered=None,
                _max_template_candidates_num=20,
            )
        else:
            self.online_template_featurizer = None
        esm_info = configs.get("esm", {})
        configs.esm.embedding_dir = f"./esm_embeddings/{configs.esm.model_name}"
        configs.esm.sequence_fpath = (
            f"./esm_embeddings/{json_task_name}_prot_sequences.csv"
        )
        self.esm_enable = esm_info.get("enable", False)
        if self.esm_enable:
            os.makedirs(configs.esm.embedding_dir, exist_ok=True)
            os.makedirs(os.path.dirname(configs.esm.sequence_fpath), exist_ok=True)
            ESMFeaturizer.precompute_esm_embedding(
                self.inputs,
                configs.esm.model_name,
                configs.esm.embedding_dir,
                configs.esm.sequence_fpath,
                configs.load_checkpoint_dir,
            )
            self.esm_featurizer = ESMFeaturizer(
                embedding_dir=esm_info.embedding_dir,
                sequence_fpath=esm_info.sequence_fpath,
                embedding_dim=esm_info.embedding_dim,
                error_dir="./esm_embeddings/",
            )

    def process_one(
        self,
        single_sample_dict: Mapping[str, Any],
    ) -> tuple[dict[str, torch.Tensor], AtomArray, dict[str, float]]:
        """
        Processes a single sample from the input JSON to generate features and statistics.

        Args:
            single_sample_dict: A dictionary containing the sample data.

        Returns:
            A tuple containing:
                - A dictionary of features.
                - An AtomArray object.
                - A dictionary of time tracking statistics.
        """
        # general features
        t0 = time.time()
        sample2feat = SampleDictToFeatures(
            single_sample_dict,
        )
        features_dict, atom_array, token_array = sample2feat.get_feature_dict()
        features_dict["distogram_rep_atom_mask"] = torch.Tensor(
            atom_array.distogram_rep_atom_mask
        ).long()
        entity_poly_type_and_seqs = (
            sample2feat.entity_poly_type_and_seqs
        )  # we include ligand as well
        t1 = time.time()
        msa_features = (
            InferenceMSAFeaturizer.make_msa_feature(
                bioassembly=single_sample_dict["sequences"],
                atom_array=atom_array,
                msa_pair_as_unpair=self.msa_pair_as_unpair,
                use_rna_msa=self.use_rna_msa,
            )
            if self.use_msa
            else {}
        )
        template_features = InferenceTemplateFeaturizer.make_template_feature(
            bioassembly=single_sample_dict["sequences"],
            atom_array=atom_array,
            use_template=self.use_template,
            online_template_featurizer=self.online_template_featurizer,
        )
        # Esm features
        if self.esm_enable:
            x_esm = self.esm_featurizer(
                token_array=token_array,
                atom_array=atom_array,
                bioassembly_dict=single_sample_dict,
                inference_mode=True,
            )
            features_dict["esm_token_embedding"] = x_esm

        # Make dummy features for not implemented features
        dummy_feats = []
        if len(template_features) == 0:
            dummy_feats.append("template")
        else:
            template_features = dict_to_tensor(template_features)
            features_dict.update(template_features)
        if len(msa_features) == 0:
            dummy_feats.append("msa")
        else:
            msa_features = dict_to_tensor(msa_features)
            features_dict.update(msa_features)
        features_dict = make_dummy_feature(
            features_dict=features_dict,
            dummy_feats=dummy_feats,
        )

        # Transform to right data type
        feat = data_type_transform(feat_or_label_dict=features_dict)

        t2 = time.time()

        data = {}
        data["input_feature_dict"] = feat
        # Add dimension related items
        N_token = feat["token_index"].shape[0]
        N_atom = feat["atom_to_token_idx"].shape[0]
        N_msa = feat["msa"].shape[0]
        stats = {}
        for mol_type in ["ligand", "protein", "dna", "rna"]:
            mol_type_mask = feat[f"is_{mol_type}"].bool()
            stats[f"{mol_type}/atom"] = int(mol_type_mask.sum(dim=-1).item())
            stats[f"{mol_type}/token"] = len(
                torch.unique(feat["atom_to_token_idx"][mol_type_mask])
            )
        N_asym = len(torch.unique(data["input_feature_dict"]["asym_id"]))
        data.update(
            {
                "N_asym": torch.tensor([N_asym]),
                "N_token": torch.tensor([N_token]),
                "N_atom": torch.tensor([N_atom]),
                "N_msa": torch.tensor([N_msa]),
            }
        )

        def formatted_key(key):
            type_, unit = key.split("/")
            if type_ == "protein":
                type_ = "prot"
            elif type_ == "ligand":
                type_ = "lig"
            else:
                pass
            return f"N_{type_}_{unit}"

        data.update(
            {
                formatted_key(k): torch.tensor([stats[k]])
                for k in [
                    "protein/atom",
                    "ligand/atom",
                    "dna/atom",
                    "rna/atom",
                    "protein/token",
                    "ligand/token",
                    "dna/token",
                    "rna/token",
                ]
            }
        )
        data.update({"entity_poly_type": entity_poly_type_and_seqs["entity_poly_type"]})
        t3 = time.time()
        time_tracker = {
            "crop": t1 - t0,
            "featurizer": t2 - t1,
            "added_feature": t3 - t2,
        }

        return data, atom_array, time_tracker

    def __len__(self) -> int:
        return len(self.inputs)

    def __getitem__(self, index: int) -> tuple[dict[str, torch.Tensor], AtomArray, str]:
        try:
            single_sample_dict = self.inputs[index]
            sample_name = single_sample_dict["name"]
            logger.info(f"Featurizing {sample_name}...")

            data, atom_array, _ = self.process_one(
                single_sample_dict=single_sample_dict
            )
            error_message = ""
        except Exception as e:
            data, atom_array = {}, None
            error_message = f"{e}:\n{traceback.format_exc()}"
        data["sample_name"] = single_sample_dict["name"]
        data["sample_index"] = index
        return data, atom_array, error_message



# Copyright 2024 ByteDance and/or its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
from typing import Any

import numpy as np
import torch
from biotite.structure import AtomArray

from protenix.data.constraint.constraint_featurizer import ConstraintFeatureGenerator
from protenix.data.core.featurizer import Featurizer
from protenix.data.core.parser import AddAtomArrayAnnot
from protenix.data.inference.json_parser import (
    add_entity_atom_array,
    remove_leaving_atoms,
)
from protenix.data.tokenizer import AtomArrayTokenizer, TokenArray
from protenix.data.utils import int_to_letters
from protenix.utils.logger import get_logger

logger = get_logger(__name__)


class SampleDictToFeatures:
    def __init__(self, single_sample_dict: dict[str, Any]) -> None:
        self.single_sample_dict = single_sample_dict
        self.input_dict = add_entity_atom_array(single_sample_dict)
        self.entity_poly_type_and_seqs = self.get_entity_poly_type_and_seqs()
        self.entity_poly_type = self.entity_poly_type_and_seqs["entity_poly_type"]
        self.entity_to_sequences = self.entity_poly_type_and_seqs["entity_to_sequences"]

    def get_entity_poly_type_and_seqs(self) -> dict[dict[str, str], dict[str, str]]:
        """
        Get the entity type for each entity.

        Allowed Value for "_entity_poly.type":
        · cyclic-pseudo-peptide
        · other
        · peptide nucleic acid
        · polydeoxyribonucleotide
        · polydeoxyribonucleotide/polyribonucleotide hybrid
        · polypeptide(D)
        · polypeptide(L)
        · polyribonucleotide

        Returns:
            dict[str, str]: a dict of polymer entity id to entity type.
        """
        entity_type_mapping_dict = {
            "proteinChain": "polypeptide(L)",
            "dnaSequence": "polydeoxyribonucleotide",
            "rnaSequence": "polyribonucleotide",
            "ligand": "non-polymer",
            "ion": "non-polymer",
        }
        entity_poly_type = {}
        entity_to_sequences = {}
        for idx, type2entity_dict in enumerate(self.input_dict["sequences"]):
            assert len(type2entity_dict) == 1, "Only one entity type is allowed."
            for entity_type, entity in type2entity_dict.items():
                if "sequence" in entity:
                    assert entity_type in [
                        "proteinChain",
                        "dnaSequence",
                        "rnaSequence",
                        "ligand",
                        "ion",
                    ], 'The "sequences" field accepts only these entity types: ["proteinChain", "dnaSequence", "rnaSequence", "ligand", "ion"].'
                    entity_poly_type[str(idx + 1)] = entity_type_mapping_dict[
                        entity_type
                    ]
                    entity_to_sequences[str(idx + 1)] = entity["sequence"]
        return {
            "entity_poly_type": entity_poly_type,
            "entity_to_sequences": entity_to_sequences,
        }

    def build_full_atom_array(self) -> AtomArray:
        """
        By assembling the AtomArray of each entity, a complete AtomArray is created.

        Returns:
            AtomArray: Biotite Atom array.
        """
        atom_array = None
        asym_chain_idx = 0
        for idx, type2entity_dict in enumerate(self.input_dict["sequences"]):
            for entity_type, entity in type2entity_dict.items():
                entity_id = str(idx + 1)

                entity_atom_array = None
                for asym_chain_count in range(1, entity["count"] + 1):
                    asym_id_str = int_to_letters(asym_chain_idx + 1)
                    asym_chain = copy.deepcopy(entity["atom_array"])
                    chain_id = [asym_id_str] * len(asym_chain)
                    copy_id = [asym_chain_count] * len(asym_chain)
                    asym_chain.set_annotation("label_asym_id", chain_id)
                    asym_chain.set_annotation("auth_asym_id", chain_id)
                    asym_chain.set_annotation("chain_id", chain_id)
                    asym_chain.set_annotation("label_seq_id", asym_chain.res_id)
                    asym_chain.set_annotation("copy_id", copy_id)
                    if entity_atom_array is None:
                        entity_atom_array = asym_chain
                    else:
                        entity_atom_array += asym_chain
                    asym_chain_idx += 1

                entity_atom_array.set_annotation(
                    "label_entity_id", [entity_id] * len(entity_atom_array)
                )

                if entity_type in ["proteinChain", "dnaSequence", "rnaSequence"]:
                    entity_atom_array.hetero[:] = False
                else:
                    entity_atom_array.hetero[:] = True

                if atom_array is None:
                    atom_array = entity_atom_array
                else:
                    atom_array += entity_atom_array
        return atom_array

    @staticmethod
    def get_a_bond_atom(
        atom_array: AtomArray,
        entity_id: int,
        position: int,
        atom_name: str,
        copy_id: int = None,
    ) -> np.ndarray:
        """
        Get the atom index of a bond atom.

        Args:
            atom_array (AtomArray): Biotite Atom array.
            entity_id (int): Entity id.
            position (int): Residue index of the atom.
            atom_name (str): Atom name.
            copy_id (copy_id): A asym chain id in N copies of an entity.

        Returns:
            np.ndarray: Array of indices for specified atoms on each asym chain.
        """
        entity_mask = atom_array.label_entity_id == str(entity_id)
        position_mask = atom_array.res_id == int(position)
        atom_name_mask = atom_array.atom_name == str(atom_name)

        if copy_id is not None:
            copy_mask = atom_array.copy_id == int(copy_id)
            mask = entity_mask & position_mask & atom_name_mask & copy_mask
        else:
            mask = entity_mask & position_mask & atom_name_mask
        atom_indices = np.where(mask)[0]
        return atom_indices

    def add_bonds_between_entities(self, atom_array: AtomArray) -> AtomArray:
        """
        Based on the information in the "covalent_bonds",
        add a bond between specified atoms on each pair of asymmetric chains of the two entities.
        Note that this requires the number of asymmetric chains in both entities to be equal.

        Args:
            atom_array (AtomArray): Biotite Atom array.

        Returns:
            AtomArray: Biotite Atom array with bonds added.
        """
        if "covalent_bonds" not in self.input_dict:
            return atom_array

        bond_count = {}
        for bond_info_dict in self.input_dict["covalent_bonds"]:
            bond_atoms = []
            for idx, i in enumerate(["left", "right"]):
                entity_id = int(
                    bond_info_dict.get(
                        f"{i}_entity", bond_info_dict.get(f"entity{idx+1}")
                    )
                )
                copy_id = bond_info_dict.get(
                    f"{i}_copy", bond_info_dict.get(f"copy{idx+1}")
                )
                position = int(
                    bond_info_dict.get(
                        f"{i}_position", bond_info_dict.get(f"position{idx+1}")
                    )
                )
                atom_name = bond_info_dict.get(
                    f"{i}_atom", bond_info_dict.get(f"atom{idx+1}")
                )

                if copy_id is not None:
                    copy_id = int(copy_id)

                if isinstance(atom_name, str):
                    if atom_name.isdigit():
                        # Convert SMILES atom index to int
                        atom_name = int(atom_name)

                if isinstance(atom_name, int):
                    # Convert AtomMap in SMILES to atom name in AtomArray
                    entity_dict = list(
                        self.input_dict["sequences"][int(entity_id - 1)].values()
                    )[0]
                    assert "atom_map_to_atom_name" in entity_dict
                    atom_name = entity_dict["atom_map_to_atom_name"][atom_name]

                # Get bond atoms by entity_id, position, atom_name
                atom_indices = self.get_a_bond_atom(
                    atom_array, entity_id, position, atom_name, copy_id
                )
                assert (
                    atom_indices.size > 0
                ), f"No atom found for {atom_name} in entity {entity_id} at position {position}."
                bond_atoms.append(atom_indices)
            assert len(bond_atoms[0]) == len(bond_atoms[1]), (
                'Can not create bonds because the "count" of entity1 '
                f'({bond_info_dict.get("left_entity", bond_info_dict.get("entity1"))}) '
                f'and entity2 ({bond_info_dict.get("right_entity", bond_info_dict.get("entity2"))}) are not equal. '
            )

            # Create bond between each asym chain pair
            for atom_idx1, atom_idx2 in zip(bond_atoms[0], bond_atoms[1]):
                atom_array.bonds.add_bond(atom_idx1, atom_idx2, 1)
                bond_count[atom_idx1] = bond_count.get(atom_idx1, 0) + 1
                bond_count[atom_idx2] = bond_count.get(atom_idx2, 0) + 1

        atom_array = remove_leaving_atoms(atom_array, bond_count)

        return atom_array

    @staticmethod
    def add_atom_array_attributes(
        atom_array: AtomArray, entity_poly_type: dict[str, str]
    ) -> AtomArray:
        """
        Add attributes to the Biotite AtomArray.

        Args:
            atom_array (AtomArray): Biotite Atom array.
            entity_poly_type (dict[str, str]): a dict of polymer entity id to entity type.

        Returns:
            AtomArray: Biotite Atom array with attributes added.
        """
        atom_array = AddAtomArrayAnnot.add_token_mol_type(atom_array, entity_poly_type)
        atom_array = AddAtomArrayAnnot.add_centre_atom_mask(atom_array)
        atom_array = AddAtomArrayAnnot.add_atom_mol_type_mask(atom_array)
        atom_array = AddAtomArrayAnnot.add_distogram_rep_atom_mask(atom_array)
        atom_array = AddAtomArrayAnnot.add_plddt_m_rep_atom_mask(atom_array)
        atom_array = AddAtomArrayAnnot.add_cano_seq_resname(atom_array)
        atom_array = AddAtomArrayAnnot.add_tokatom_idx(atom_array)
        atom_array = AddAtomArrayAnnot.add_modified_res_mask(atom_array)
        atom_array = AddAtomArrayAnnot.unique_chain_and_add_ids(atom_array)
        atom_array = AddAtomArrayAnnot.find_equiv_mol_and_assign_ids(
            atom_array, check_final_equiv=False
        )
        atom_array = AddAtomArrayAnnot.add_ref_space_uid(atom_array)
        return atom_array

    @staticmethod
    def mse_to_met(atom_array: AtomArray) -> AtomArray:
        """
        Ref: AlphaFold3 SI chapter 2.1
        MSE residues are converted to MET residues.

        Args:
            atom_array (AtomArray): Biotite AtomArray object.

        Returns:
            AtomArray: Biotite AtomArray object after converted MSE to MET.
        """
        mse = atom_array.res_name == "MSE"
        se = mse & (atom_array.atom_name == "SE")
        atom_array.atom_name[se] = "SD"
        atom_array.element[se] = "S"
        atom_array.res_name[mse] = "MET"
        atom_array.hetero[mse] = False
        return atom_array

    def get_atom_array(self) -> AtomArray:
        """
        Create a Biotite AtomArray and add attributes from the input dict.

        Returns:
            AtomArray: Biotite Atom array.
        """
        atom_array = self.build_full_atom_array()
        atom_array = self.add_bonds_between_entities(atom_array)
        atom_array = self.mse_to_met(atom_array)
        atom_array = self.add_atom_array_attributes(atom_array, self.entity_poly_type)
        return atom_array

    def get_feature_dict(self) -> tuple[dict[str, torch.Tensor], AtomArray, TokenArray]:
        """
        Generates a feature dictionary from the input sample dictionary.

        Returns:
            A tuple containing:
                - A dictionary of features.
                - An AtomArray object.
                - A TokenArray object.
        """
        atom_array = self.get_atom_array()

        aa_tokenizer = AtomArrayTokenizer(atom_array)
        token_array = aa_tokenizer.get_token_array()

        (
            constraint_feature,
            token_array,
            atom_array,
        ) = ConstraintFeatureGenerator.generate_from_json(
            token_array,
            atom_array,
            self.input_dict["sequences"],
            self.input_dict.get("constraint", {}),
        )

        featurizer = Featurizer(
            token_array, atom_array, include_discont_poly_poly_bonds=True
        )
        feature_dict = featurizer.get_all_input_features()
        feature_dict["constraint_feature"] = constraint_feature

        token_array_with_frame = featurizer.get_token_frame(
            token_array=token_array,
            atom_array=atom_array,
            ref_pos=feature_dict["ref_pos"],
            ref_mask=feature_dict["ref_mask"],
        )

        # [N_token]
        feature_dict["has_frame"] = torch.Tensor(
            token_array_with_frame.get_annotation("has_frame")
        ).long()

        # [N_token, 3]
        feature_dict["frame_atom_index"] = torch.Tensor(
            token_array_with_frame.get_annotation("frame_atom_index")
        ).long()
        return feature_dict, atom_array, token_array



# Copyright 2024 ByteDance and/or its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional

import torch


def rmsd(
    pred_pose: torch.Tensor,
    true_pose: torch.Tensor,
    mask: torch.Tensor = None,
    eps: float = 0.0,
    reduce: bool = True,
):
    """
    compute rmsd between two poses, with the same shape
    Arguments:
        pred_pose/true_pose: [...,N,3], two poses with the same shape
        mask: [..., N], mask to indicate which atoms/pseudo_betas/etc to compute
        eps: add a tolerance to avoid floating number issue
        reduce: decide the return shape of rmsd;
    Return:
        rmsd: if reduce = true, return the mean of rmsd over batches;
            else return a tensor containing each rmsd separately
    """

    # mask [..., N]
    assert pred_pose.shape == true_pose.shape  # [..., N, 3]

    if mask is None:
        mask = torch.ones(true_pose.shape[:-1], device=true_pose.device)

    # [...]
    err2 = (torch.square(pred_pose - true_pose).sum(dim=-1) * mask).sum(
        dim=-1
    ) / mask.sum(dim=-1)
    rmsd = err2.add(eps).sqrt()
    if reduce:
        rmsd = rmsd.mean()
    return rmsd


def align_pred_to_true(
    pred_pose: torch.Tensor,
    true_pose: torch.Tensor,
    atom_mask: Optional[torch.Tensor] = None,
    weight: Optional[torch.Tensor] = None,
    allowing_reflection: bool = False,
):
    """Find optimal transformation, rotation (and reflection) of two poses.
    Arguments:
        pred_pose: [...,N,3] the pose to perform transformation on
        true_pose: [...,N,3] the target pose to align pred_pose to
        atom_mask: [..., N] a mask for atoms
        weight: [..., N] a weight vector to be applied.
        allow_reflection: whether to allow reflection when finding optimal alignment
    return:
        aligned_pose: [...,N,3] the transformed pose
        rot: optimal rotation
        translate: optimal translation
    """
    if atom_mask is not None:
        pred_pose = pred_pose * atom_mask.unsqueeze(-1)
        true_pose = true_pose * atom_mask.unsqueeze(-1)
    else:
        atom_mask = torch.ones(*pred_pose.shape[:-1]).to(pred_pose.device)

    if weight is None:
        weight = atom_mask
    else:
        weight = weight * atom_mask

    weighted_n_atoms = torch.sum(weight, dim=-1, keepdim=True).unsqueeze(-1)
    pred_pose_centroid = (
        torch.sum(pred_pose * weight.unsqueeze(-1), dim=-2, keepdim=True)
        / weighted_n_atoms
    )
    pred_pose_centered = pred_pose - pred_pose_centroid
    true_pose_centroid = (
        torch.sum(true_pose * weight.unsqueeze(-1), dim=-2, keepdim=True)
        / weighted_n_atoms
    )
    true_pose_centered = true_pose - true_pose_centroid
    H_mat = torch.matmul(
        (pred_pose_centered * weight.unsqueeze(-1)).transpose(-2, -1),
        true_pose_centered * atom_mask.unsqueeze(-1),
    )
    u, s, vh = torch.linalg.svd(H_mat)
    u = u.transpose(-1, -2)
    v = vh.transpose(-1, -2)

    if not allowing_reflection:

        det = torch.linalg.det(torch.matmul(v, u))

        diagonal = torch.stack(
            [torch.ones_like(det), torch.ones_like(det), det], dim=-1
        )
        rot = torch.matmul(
            torch.diag_embed(diagonal).to(u.device),
            u,
        )
        rot = torch.matmul(v, rot)
    else:
        rot = torch.matmul(v, u)
    translate = true_pose_centroid - torch.matmul(
        pred_pose_centroid, rot.transpose(-1, -2)
    )

    pred_pose_translated = (
        torch.matmul(pred_pose_centered, rot.transpose(-1, -2)) + true_pose_centroid
    )

    return pred_pose_translated, rot, translate


def partially_aligned_rmsd(
    pred_pose: torch.Tensor,
    true_pose: torch.Tensor,
    align_mask: torch.Tensor,
    atom_mask: torch.Tensor,
    weight: Optional[torch.Tensor] = None,
    eps: float = 0.0,
    reduce: bool = True,
    allowing_reflection: bool = False,
):
    """RMSD when aligning parts of the complex coordinate, does NOT take permutation symmetricity into consideration
    Arguments:
        pred_pose: native predicted pose, [..., N,3]
        true_pose: ground truth pose, [..., N, 3]
        align_mask: a mask representing which coordinates to align [..., N]
        atom_mask: a mask representing which coordinates to compute loss [..., N]
        weight: a weight tensor assining weights in alignment for each atom [..., N]
        eps: add a tolerance to avoid floating number issue in sqrt
        reduce: decide the return shape of rmsd;
        allowing_reflection: whether to allow reflection when finding optimal alignment
    return:
        aligned_part_rmsd: the rmsd of part being align_masked
        unaligned_part_rmsd: the rmsd of unaligned part
        transformed_pred_pose:
        rot: optimal rotation
        trans: optimal translation
    """
    _, rot, translate = align_pred_to_true(
        pred_pose,
        true_pose,
        atom_mask=atom_mask * align_mask,
        weight=weight,
        allowing_reflection=allowing_reflection,
    )
    transformed_pose = torch.matmul(pred_pose, rot.transpose(-1, -2)) + translate
    err_atom = torch.square(transformed_pose - true_pose).sum(dim=-1) * atom_mask
    aligned_mask, unaligned_mask = atom_mask * align_mask.float(), atom_mask * (
        1 - align_mask.float()
    )
    aligned_part_err_square = (err_atom * aligned_mask).sum(dim=-1) / aligned_mask.sum(
        dim=-1
    )
    unaligned_part_err_square = (err_atom * unaligned_mask).sum(
        dim=-1
    ) / unaligned_mask.sum(dim=-1)
    aligned_part_rmsd = aligned_part_err_square.add(eps).sqrt()
    unaligned_part_rmsd = unaligned_part_err_square.add(eps).sqrt()
    if reduce:
        aligned_part_rmsd = aligned_part_rmsd.mean()
        unaligned_part_rmsd = unaligned_part_rmsd.mean()
    return aligned_part_rmsd, unaligned_part_rmsd, transformed_pose, rot, translate


def self_aligned_rmsd(
    pred_pose: torch.Tensor,
    true_pose: torch.Tensor,
    atom_mask: torch.Tensor,
    eps: float = 0.0,
    reduce: bool = True,
    allowing_reflection: bool = False,
):
    """RMSD when aligning one molecule with ground truth and compute rmsd.
    Arguments:
        pred_pose: native predicted pose, [..., N,3]
        true_pose: ground truth pose, [..., N, 3]
        atom_mask: a mask representing which coordinates to compute loss [..., N]
        eps: add a tolerance to avoid floating number issue in sqrt
        reduce: decide the return shape of rmsd;
        allowing_reflection: whether to allow reflection when finding optimal alignment
    return:
        aligned_rmsd: the rmsd of part being align_masked
        transformed_pred_pose: the aligned pose
        rot: optimal rotation matrix
        trans: optimal translation
    """
    aligned_rmsd, _, transformed_pred_pose, rot, trans = partially_aligned_rmsd(
        pred_pose=pred_pose,
        true_pose=true_pose,
        align_mask=atom_mask,
        atom_mask=atom_mask,
        eps=eps,
        reduce=reduce,
        allowing_reflection=allowing_reflection,
    )
    return aligned_rmsd, transformed_pred_pose, rot, trans


def weighted_rigid_align(
    x: torch.Tensor,
    x_target: torch.Tensor,
    atom_weight: torch.Tensor,
    stop_gradient: bool = True,
) -> tuple[torch.Tensor]:
    """Implements Algorithm 28 in AF3. Wrap `align_pred_to_true`.

    Args:
        x (torch.Tensor): input coordinates, it will be moved to match x_target.
            [..., N_atom, 3]
        x_target (torch.Tensor): target coordinates for the input to match.
            [..., N_atom, 3]
        atom_weight (torch.Tensor): weights for each atom.
            [..., N_atom] or [N_atom]
        stop_gradient (bool): whether to detach the output. If true, will run it with no_grad() ctx.

    Returns:
        x_aligned (torch.Tensor): rotated, translated x which should be closer to x_target.
            [..., N_atom, 3]
    """

    if len(atom_weight.shape) == len(x.shape) - 1:
        assert atom_weight.shape[:-1] == x.shape[:-2]
    else:
        assert len(atom_weight.shape) == 1 and atom_weight.shape[-1] == x.shape[-2]

    if stop_gradient:
        with torch.no_grad():
            x_aligned, rot, trans = align_pred_to_true(
                pred_pose=x,
                true_pose=x_target,
                atom_mask=None,
                weight=atom_weight,
                allowing_reflection=False,
            )
            return x_aligned.detach()
    else:
        x_aligned, rot, trans = align_pred_to_true(
            pred_pose=x,
            true_pose=x_target,
            atom_mask=None,
            weight=atom_weight,
            allowing_reflection=False,
        )
        return x_aligned



# Copyright 2024 ByteDance and/or its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import time

import torch


class Checker:
    @staticmethod
    def is_permutation(x: torch.Tensor):
        """
        Checks if the input tensor `x` is a permutation of integers from 0 to N-1.

        Args:
            x (torch.Tensor): A 1D tensor of size [N].
        """
        assert x.dim() == 1
        N = x.size(0)
        assert torch.equal(torch.sort(x)[0], torch.arange(N, device=x.device))

    @staticmethod
    def are_permutations(x: torch.Tensor, dim=-1):
        """
        Checks if slices along the specified dimension in `x` are permutations of integers from 0 to N-1.

        Args:
            x (torch.Tensor): A tensor with any number of dimensions, containing slices of size N along `dim`.
            dim (int, optional): The dimension along which to check for permutations. Defaults to -1.
        """
        assert x.dim() > 0

        N = x.size(dim)
        # Create a view of x that moves the specified dimension to -1
        x = x.transpose(dim, -1).contiguous()
        x = x.reshape(-1, N)
        expected = torch.arange(N, device=x.device)
        for i in range(x.size(0)):
            Checker.is_permutation(x[i])

    @staticmethod
    def contains_identity(x: torch.Tensor, dim=-1):
        """
        Check if x contains the identity permutation

        Args:
            x (torch.Tensor): A tensor with any number of dimensions, containing slices of size N along `dim`.
            dim (int, optional): The dimension along which to check for permutations. Defaults to -1.
        """
        assert x.dim() > 0

        N = x.size(dim)
        # Create a view of x that moves the specified dimension to -1
        x = x.transpose(dim, -1).contiguous()
        x = x.reshape(-1, N)
        expected = torch.arange(N, device=x.device).unsqueeze(dim=0)
        assert (x == expected).all(dim=-1).any()

    @staticmethod
    def not_contain_identity(x: torch.Tensor, dim=-1):
        """
        Check if x does not contain the identity permutation

        Args:
            x (torch.Tensor): A tensor with any number of dimensions, containing slices of size N along `dim`.
            dim (int, optional): The dimension along which to check for permutations. Defaults to -1.
        """
        assert x.dim() > 0

        N = x.size(dim)
        # Create a view of x that moves the specified dimension to -1
        x = x.transpose(dim, -1).contiguous()
        x = x.reshape(-1, N)
        expected = torch.arange(N, device=x.device).unsqueeze(dim=0)
        assert not (x == expected).all(dim=-1).any()

    @staticmethod
    def batch_permute(perm: torch.Tensor, x: torch.Tensor, x_permuted: torch.Tensor):
        """
        Args:
            perm (torch.Tensor):
                [..., N]
            x (torch.Tensor):
                [N, batch_dims_x]
            x_permuted (torch.Tensor):
                [..., N, batch_dims_x]
        """
        batch_shape = perm.shape[:-1]
        N = perm.size(-1)
        assert x.size(0) == N
        perm = perm.view(-1, N)
        permuted_x = [x[perm[i]] for i in range(len(perm))]
        permuted_x = torch.stack(permuted_x, dim=0)  # [-1, N, batch_dims_x]
        target_shape = batch_shape + (N,) + x.shape[1:]
        assert torch.allclose(permuted_x.reshape(target_shape), x_permuted)


def save_permutation_error(data, error_dir: str = None, max_cases: int = 50):
    """
    Saves the permutation error data to a specified directory.

    Args:
        data: The data to be saved.
        error_dir (str): The directory where the error data should be saved.
        max_cases (int): The maximum number of error cases to save.

    Raises:
        Exception: If an error occurs while saving the data, the exception is caught and printed.
    """
    if error_dir is None:
        return

    # error_dir = os.path.join(self.error_dir, dir_name)
    os.makedirs(error_dir, exist_ok=True)

    if len(os.listdir(error_dir)) >= max_cases:
        # Only record the first {max_cases} error cases for debug
        return

    filename = "T_" + time.strftime("%Y%m%d_%H%M%S") + ".pt"
    fpath = os.path.join(error_dir, filename)
    if not os.path.exists(fpath):
        try:
            torch.save(data, fpath)
        except Exception as e:
            print(f"Exception occurrs in save_permutation_error: {e}")



# Copyright 2024 ByteDance and/or its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from contextlib import nullcontext
from typing import Sequence, Union

import numpy as np
import torch
from torch import nn
from torch.nn.parameter import Parameter


def grad_norm(params):
    total_norm = 0.0
    for p in params:
        if p.grad is None:
            continue
        param_norm = p.grad.data.norm(2)
        total_norm += param_norm.item() ** 2
    total_norm = total_norm ** (1.0 / 2)
    return total_norm


def to_device(obj, device):
    """Move tensor or dict of tensors to device"""
    if isinstance(obj, dict):
        for k, v in obj.items():
            if isinstance(v, dict):
                to_device(v, device)
            elif isinstance(v, torch.Tensor):
                obj[k] = obj[k].to(device)
    elif isinstance(obj, torch.Tensor):
        obj = obj.to(device)
    else:
        raise Exception(f"type {type(obj)} not supported")
    return obj


def detach_if(t: torch.Tensor, detach: bool):
    if detach:
        return t.detach()
    return t


def cdist(a: torch.Tensor, b: torch.Tensor = None):
    # for tensor shape [1, 512 * 14, 3], donot_use_mm_for_euclid_dist mode costs 0.0489s,
    # while use_mm_for_euclid_dist_if_necessary costs 0.0419s on cpu. On GPU there two costs
    # will be neglectible. So there is no need to sacrifice accuracy for speed here.
    return torch.cdist(
        a,
        b if b is not None else a,
        compute_mode="donot_use_mm_for_euclid_dist",
    )


def batch_avg_with_mask(
    value: torch.Tensor,
    mask: torch.Tensor,
    avg_dim: Union[int, tuple[int]] = None,
    batch_reduction: str = "mean",
    eps: float = 1e-12,
):
    """Average values with mask.
    Args:
        value: tensor of shape [BS, ...]
        mask: tensor with same shape and type of value, 1 means valid, 0 means maksed
        avg_dim: dimensions to apply average, if None, all dims excluding BS dim will be averaged
        batch_reduction: mean/sum/none, reduction operation applied on BS dim
    """
    if avg_dim is None:
        avg_dim = tuple(range(1, len(value.shape)))
    avg = (value * mask).sum(dim=avg_dim) / (mask.sum(dim=avg_dim) + eps)
    if batch_reduction == "mean":
        return avg.mean()
    elif batch_reduction == "sum":
        return avg.sum()
    elif batch_reduction == "none":
        return avg
    else:
        raise Exception(f"Invalid batch_reduction: {batch_reduction}")


def eye_mask(L, device=None, opposite=False):
    if opposite:
        return 1.0 - torch.eye(L, device=device)
    else:
        torch.eye(L, device=device)


def glorot_uniform(t):
    if len(t.size()) == 2:
        fan_in, fan_out = t.size()
    elif len(t.size()) == 3:
        # out_ch, in_ch, kernel for Conv 1
        fan_in = t.size()[1] * t.size()[2]
        fan_out = t.size()[0] * t.size()[2]
    else:
        fan_in = np.prod(t.size())
        fan_out = np.prod(t.size())

    limit = np.sqrt(6.0 / (fan_in + fan_out))
    t.uniform_(-limit, limit)


def _param_init(m, bias="zero"):
    if isinstance(m, Parameter):
        glorot_uniform(m.data)
    elif isinstance(m, nn.Linear):
        if m.bias is not None:
            if bias == "zero":
                m.bias.data.zero_()
            else:
                assert bias == "normal"
                m.bias.data.normal_()
        glorot_uniform(m.weight.data)


def weights_init(m, bias="zero"):
    for p in m.modules():
        if isinstance(p, nn.ParameterList) or isinstance(p, nn.ModuleList):
            for pp in p:
                _param_init(pp, bias)
        else:
            _param_init(p, bias)

    for name, p in m.named_parameters():
        if "." not in name:  # top-level parameters
            _param_init(p, bias)


def permute_last_dims(t: torch.Tensor, dims: Sequence[int]):
    """Permute tensor on last dims, all other dims are kept unchanged.

    Args:
        t (torch.Tensor): Input tensor with at least len(dims) dimensions.
        dims: The desired ordering of dimensions, here all values should be < 0, i.e. (-1, -2) means permute last two dims.
    """
    num_dims = len(t.shape)
    prefix_dims = list(range(num_dims - len(dims)))
    last_dims = [num_dims + d for d in dims]
    return torch.permute(t, prefix_dims + last_dims)


def flatten_tensors(tensors) -> torch.Tensor:
    """Flatten a list of tensors into a single 1D tensor."""
    return torch.cat([t.view(-1) for t in tensors], dim=0)


def unflatten_tensors(flat_tensor, shapes):
    """Unflatten a 1D tensor into a list of tensors with given shapes."""
    tensors = []
    offset = 0
    for shape in shapes:
        numel = shape.numel()
        tensors.append(flat_tensor[offset : offset + numel].view(shape))
        offset += numel
    return tensors


def map_values_to_list(data, recursive=True):
    for k, v in data.items():
        if isinstance(v, torch.Tensor):
            if v.dtype == torch.bfloat16:
                v = v.float()
            data[k] = v.cpu().numpy().tolist()
        elif isinstance(v, np.ndarray):
            data[k] = v.tolist()
        elif isinstance(v, dict) and recursive:
            data[k] = map_values_to_list(v, recursive)
    return data


def round_values(data, recursive=True):
    for k, v in data.items():
        if isinstance(v, torch.Tensor):
            if v.dtype == torch.bfloat16:
                v = v.float()
            data[k] = np.round(v.cpu().numpy(), 2)
        elif isinstance(v, np.ndarray):
            data[k] = np.round(v, 2)
        elif isinstance(v, list):
            data[k] = list(np.round(np.array(v), 2))
        elif isinstance(v, dict) and recursive:
            data[k] = round_values(v, recursive)
    return data


def autocasting_disable_decorator(disable_casting):
    def func_wrapper(func):
        def new_func(*args, **kwargs):
            _amp_context = (
                torch.autocast(device_type="cuda", enabled=False)
                if disable_casting
                else nullcontext()
            )

            # Helper function to conditionally cast tensors
            def conditioned_cast(tensor):
                if (
                    disable_casting
                    and isinstance(tensor, torch.Tensor)
                    and torch.is_floating_point(tensor)
                ):
                    return tensor.to(dtype=torch.float32)
                return tensor

            with _amp_context:
                return func(
                    *(conditioned_cast(v) for v in args),
                    **{k: conditioned_cast(v) for k, v in kwargs.items()},
                )

        return new_func

    return func_wrapper


def dict_to_tensor(feature_dict):
    for k, v in feature_dict.items():
        if not isinstance(v, torch.Tensor):
            dtype = feature_dict[k].dtype
            feature_dict[k] = torch.tensor(v)

            if dtype in [np.int64, np.int32]:
                feature_dict[k] = feature_dict[k].to(torch.int64)
            elif dtype in [np.float32, np.float64]:
                feature_dict[k] = feature_dict[k].to(torch.float32)

    return feature_dict


def collate_fn_identity(x):
    return x


def collate_fn_first(x):
    return x[0]
